"""_forestry.py: functions to support management of parts in Model."""

import logging

log = logging.getLogger(__name__)

import copy
import os
import shutil
import zipfile as zf

import resqpy.olio.consolidation as cons
import resqpy.olio.uuid as bu
import resqpy.olio.write_hdf5 as whdf5
import resqpy.olio.xml_et as rqet
from resqpy.olio.xml_namespaces import curly_namespace as ns
from resqpy.olio.xml_namespaces import namespace as ns_url


def _load_part(model, epc, part_name, is_rels = None):
    """Load and parse xml tree for given part name, storing info in parts forest (or rels forest)."""

    # note: epc is 'open' ZipFile handle
    if part_name.startswith('/'):
        part_name = part_name[1:]

    try:

        # log.debug('loading part ' + part_name)
        if is_rels is None:
            is_rels = part_name.endswith('.rels')
        is_other = not is_rels and part_name.startswith('docProps')

        part_type = None
        part_uuid = None

        if is_rels:
            if part_name in model.rels_forest:
                (part_uuid, _) = model.rels_forest[part_name]
            if part_uuid is None:
                part_uuid = rqet.uuid_in_part_name(part_name)
        elif is_other:
            (part_type, _) = model.other_forest[part_name]
        else:
            (part_type, part_uuid, _) = model.parts_forest[part_name]  # part_type must already have been established

        with epc.open(part_name) as part_xml:
            part_tree = rqet.parse(part_xml)
            if is_rels:
                model.rels_forest[part_name] = (part_uuid, part_tree)
            elif is_other:
                model.other_forest[part_name] = (part_type, part_tree)
            else:
                uuid_from_tree = rqet.uuid_for_part_root(part_tree.getroot())
                if part_uuid is None:
                    part_uuid = uuid_from_tree
                elif uuid_from_tree is not None:
                    assert bu.matching_uuids(part_uuid, uuid_from_tree)
                model.parts_forest[part_name] = (part_type, part_uuid, part_tree)
                _set_uuid_to_part(model, part_name)
                if model.crs_uuid is None and part_type == 'obj_LocalDepth3dCrs':  # randomly assign first crs as primary crs for model
                    model.crs_uuid = part_uuid

        return True

    except Exception:

        log.exception('(okay to continue?) failed to load part: ' + part_name)
        return False


def _fell_part(model, part_name):
    """Removes the named part from the in-memory parts forest."""

    try:
        _del_uuid_to_part(model, part_name)
    except Exception:
        pass
    try:
        del model.parts_forest[part_name]
    except Exception:
        pass
    try:
        del model.rels_forest[part_name]
    except Exception:
        pass
    try:
        del model.other_forest[part_name]
    except Exception:
        pass


def _remove_part_from_main_tree(model, part):
    """Removes the named part from the main (Content_Types) tree."""

    for child in model.main_root:
        if rqet.stripped_of_prefix(child.tag) == 'Override':
            part_name = child.attrib['PartName']
            if part_name[0] == '/':
                part_name = part_name[1:]
            if part_name == part:
                log.debug('removing part from main xml tree: ' + part)
                model.main_root.remove(child)
                break


def _tidy_up_forests(model, tidy_main_tree = True, tidy_others = False, remove_extended_core = True):
    """Removes any parts that do not have any related data in dictionaries."""

    deletion_list = []
    for part, info in model.parts_forest.items():
        if info == (None, None, None):
            deletion_list.append(part)
    for part in deletion_list:
        log.debug('removing part due to lack of xml tree etc.: ' + str(part))
        if tidy_main_tree:
            _remove_part_from_main_tree(model, part)
        _del_uuid_to_part(model, part)
        del model.parts_forest[part]
    deletion_list = []
    for part, info in model.rels_forest.items():
        if info == (None, None):
            deletion_list.append(part)
    for part in deletion_list:
        log.debug('removing rels part due to lack of xml tree etc.: ' + str(part))
        if tidy_main_tree:
            _remove_part_from_main_tree(model, part)
        del model.rels_forest[part]
    if tidy_others:
        for part, info in model.other_forest.items():
            if info == (None, None):
                deletion_list.append(part)
        for part in deletion_list:
            log.debug('removing docProps part due to lack of xml tree etc.: ' + str(part))
            if tidy_main_tree:
                _remove_part_from_main_tree(model, part)
            del model.other_forest[part]
    if remove_extended_core and 'docProps/extendedCore.xml' in model.other_forest:  # more trouble than it's worth
        part = 'docProps/extendedCore.xml'
        if tidy_main_tree:
            _remove_part_from_main_tree(model, part)
        del model.other_forest[part]


def _load_epc(model, epc_file, full_load = True, epc_subdir = None, copy_from = None):
    """Load xml parts of model from epc file (HDF5 arrays are not loaded)."""

    if not epc_file.endswith('.epc'):
        epc_file += '.epc'

    _copy_dataset_if_requested(copy_from, epc_file)

    log.info('loading resqml model from epc file ' + epc_file)

    if model.modified:
        log.warning('loading model from epc, discarding previous in-memory modifications')
        model.initialize()

    model.set_epc_file_and_directory(epc_file)

    with zf.ZipFile(epc_file) as epc:
        names = epc.namelist()
        _set_part_names_in_forests(model, epc_subdir, names)
        with epc.open('[Content_Types].xml') as main_xml:
            model.main_tree = rqet.parse(main_xml)
            model.main_root = model.main_tree.getroot()
            for child in model.main_root:
                _complete_forest_entry_for_part(epc, model, epc_subdir, full_load, child)
        if model.rels_present and full_load:
            _load_relationships(epc, model, epc_subdir, names)
            if copy_from:
                model.change_filename_in_hdf5_rels(os.path.split(epc_file)[1][:-4] + '.h5')
        elif not model.rels_present:
            assert len(model.rels_forest) == 0
        if full_load:
            _tidy_up_forests(model)


def _copy_dataset_if_requested(copy_from, epc_file):
    if copy_from:
        if not copy_from.endswith('.epc'):
            copy_from += '.epc'
        log.info('copying ' + copy_from + ' to ' + epc_file + ' along with paired .h5 files')
        shutil.copy(copy_from, epc_file)
        shutil.copy(copy_from[:-4] + '.h5', epc_file[:-4] + '.h5')


def _exclude(name, epc_subdir):
    if epc_subdir is None:
        return False
    if '/' not in name:
        return False
    if name.startswith('docProps') or name.startswith('_rels'):
        return False
    if isinstance(epc_subdir, str):
        epc_subdir = [epc_subdir]
    for subdir in epc_subdir:
        if subdir.endswith('/'):
            head = subdir
        else:
            head = subdir + '/'
        if name.startswith(head):
            return False
    return True


def _set_part_names_in_forests(model, epc_subdir, names):
    for name in names:
        if _exclude(name, epc_subdir):
            continue
        if name != '[Content_Types].xml':
            if name.startswith('docProps'):
                model.other_forest[name] = (None, None)  # used for non-uuid parts, ie. docProps
            else:
                part_uuid = rqet.uuid_in_part_name(name)
                if '_rels' in name:
                    model.rels_forest[name] = (part_uuid, None)
                else:
                    model.parts_forest[name] = (None, part_uuid, None)
                    _set_uuid_to_part(model, name)


def _complete_forest_entry_for_part(epc, model, epc_subdir, full_load, child):
    if rqet.stripped_of_prefix(child.tag) == 'Override':
        attrib_dict = child.attrib
        part_name = attrib_dict['PartName']
        if part_name[0] == '/':
            part_name = part_name[1:]
        part_type = rqet.content_type(attrib_dict['ContentType'])
        if part_name.startswith('docProps'):
            if part_name not in model.other_forest:
                log.warning('docProps entry in Content_Types does not exist as part in epc: ' + part_name)
                return
            model.other_forest[part_name] = (part_type, None)
        else:
            if part_name not in model.parts_forest:
                if epc_subdir is None:
                    log.warning('entry in Content_Types does not exist as part in epc: ' + part_name)
                return
            part_uuid = model.parts_forest[part_name][1]
            model.parts_forest[part_name] = (part_type, part_uuid, None)
        if full_load:
            load_success = _load_part(model, epc, part_name)
            if not load_success:
                _fell_part(model, part_name)
    elif rqet.stripped_of_prefix(child.tag) == 'Default':
        if 'Extension' in child.attrib.keys() and child.attrib['Extension'] == 'rels':
            assert not model.rels_present
            model.rels_present = True
    else:
        # todo: check standard for other valid tags
        pass


def _load_relationships(epc, model, epc_subdir, names):
    for name in names:
        if _exclude(name, epc_subdir):
            continue
        if name.startswith('_rels/'):
            load_success = _load_part(model, epc, name, is_rels = True)
            if not load_success:
                _fell_part(model, name)


def _store_epc(model, epc_file = None, main_xml_name = '[Content_Types].xml', only_if_modified = False):
    """Write xml parts of model to epc file (HDF5 arrays are not written here)."""

    # for prefix, uri in ns.items():
    #     et.register_namespace(prefix, uri)

    if not epc_file:
        epc_file = model.epc_file
    assert epc_file, 'no file name given or known when attempting to store epc'

    if only_if_modified and not model.modified:
        return

    log.info('storing resqml model to epc file ' + epc_file)

    assert model.main_tree is not None
    if model.main_root is None:
        model.main_root = model.main_tree.getroot()

    with zf.ZipFile(epc_file, mode = 'w') as epc:
        with epc.open(main_xml_name, mode = 'w') as main_xml:
            log.debug('Writing main xml: ' + main_xml_name)
            rqet.write_xml(main_xml, model.main_tree, standalone = 'yes')
        for part_name, (_, _, part_tree) in model.parts_forest.items():
            if part_tree is None:
                log.warning('No xml tree present to write for part: ' + part_name)
                continue
            if part_name[0] == '/':
                part_name = part_name[1:]
            with epc.open(part_name, mode = 'w') as part_xml:
                rqet.write_xml(part_xml, part_tree, standalone = None)
        for part_name, (_, part_tree) in model.other_forest.items():
            if part_tree is None:
                log.warning('No xml tree present to write for other part: ' + part_name)
                continue
            if part_name[0] == '/':
                part_name = part_name[1:]
            with epc.open(part_name, mode = 'w') as part_xml:
                rqet.write_xml(part_xml, part_tree, standalone = 'yes')
        if model.rels_present:
            for part_name, (_, part_tree) in model.rels_forest.items():
                if part_tree is None:
                    log.warning('No xml tree present to write for rels part: ' + part_name)
                    continue
                with epc.open(part_name, mode = 'w') as part_xml:
                    rqet.write_xml(part_xml, part_tree, standalone = 'yes')
        # todo: other parts (documentation etc.)
    model.set_epc_file_and_directory(epc_file)
    model.modified = False


def _copy_part(model, existing_uuid, new_uuid, change_hdf5_refs = False):
    """Makes a new part as a copy of an existing part with only a new uuid set; the new part can then be modified."""

    old_uuid_str = str(existing_uuid)
    new_uuid_str = str(new_uuid)
    log.debug('copying xml part from uuid: ' + old_uuid_str + ' to uuid: ' + new_uuid_str)
    existing_parts_list = model.parts_list_of_type(uuid = existing_uuid)
    if len(existing_parts_list) == 0:
        log.warning('failed to find existing part for copying with uuid: ' + old_uuid_str)
        return None
    assert len(existing_parts_list) == 1, 'more than one existing part found with uuid: ' + old_uuid_str
    (part_type, old_uuid, old_tree) = model.parts_forest[existing_parts_list[0]]
    assert bu.matching_uuids(old_uuid, existing_uuid)
    new_tree = copy.deepcopy(old_tree)
    new_root = new_tree.getroot()
    part_name = rqet.patch_uuid_in_part_root(new_root, new_uuid)
    if change_hdf5_refs:
        model.change_uuid_in_hdf5_references(new_root, old_uuid_str, new_uuid_str)
    _add_part(model, part_type, new_uuid, new_root, add_relationship_part = False)
    return part_name


def _add_part(model, content_type, uuid, root, add_relationship_part = True, epc_subdir = None):
    """Adds a (recently created) node as a new part in the model's parts forest."""

    if content_type[0].isupper():
        content_type = 'obj_' + content_type
    use_other = (content_type == 'docProps')
    if use_other:
        if rqet.pretend_to_be_fesapi or rqet.use_fesapi_quirks:
            prefix = '/'
        else:
            prefix = ''
        part_name = prefix + 'docProps/core.xml'
        ct = 'application/vnd.openxmlformats-package.core-properties+xml'
    else:
        part_name = rqet.part_name_for_object(content_type, uuid, prefixed = False, epc_subdir = epc_subdir)
        if 'EpcExternalPartReference' in content_type:
            ct = 'application/x-eml+xml;version=2.0;type=' + content_type
        else:
            ct = 'application/x-resqml+xml;version=2.0;type=' + content_type

    # log.debug('adding part: ' + part_name)
    if isinstance(uuid, str):
        uuid = bu.uuid_from_string(uuid)
    part_tree = rqet.ElementTree(element = root)
    if use_other:
        model.other_forest[part_name] = (content_type, part_tree)
    else:
        model.parts_forest[part_name] = (content_type, uuid, part_tree)
        _set_uuid_to_part(model, part_name)
    main_ref = rqet.SubElement(model.main_root, ns['content_types'] + 'Override')
    main_ref.set('PartName', part_name)
    main_ref.set('ContentType', ct)
    if add_relationship_part and model.rels_present:
        rels_node = rqet.Element(ns['rels'] + 'Relationships')
        rels_node.text = '\n'
        rels_tree = rqet.ElementTree(element = rels_node)
        if use_other:
            rels_part_name = '_rels/.rels'
        else:
            rels_part_name = rqet.rels_part_name_for_part(part_name)
        model.rels_forest[rels_part_name] = (uuid, rels_tree)
    model.set_modified()


def _patch_root_for_part(model, part, root):
    """Updates the xml tree for the part without changing the uuid."""

    content_type, uuid, part_tree = model.parts_forest[part]
    assert bu.matching_uuids(uuid, rqet.uuid_for_part_root(root))
    part_tree = rqet.ElementTree(element = root)
    model.parts_forest[part] = (content_type, uuid, part_tree)


def _remove_part(model, part_name, remove_relationship_part):
    """Removes a part from the parts forest; optionally remove corresponding rels part and other relationships."""

    _del_uuid_to_part(model, part_name)
    model.parts_forest.pop(part_name)
    if remove_relationship_part:
        if 'docProps' in part_name:
            rels_part_name = '_rels/.rels'
        else:
            related_parts = model.parts_list_filtered_by_related_uuid(model.list_of_parts(),
                                                                      rqet.uuid_in_part_name(part_name))
            for relative in related_parts:
                (rel_uuid, rel_tree) = model.rels_forest[rqet.rels_part_name_for_part(relative)]
                rel_root = rel_tree.getroot()
                for child in rel_root:
                    if rqet.stripped_of_prefix(child.tag) != 'Relationship':
                        continue
                    if child.attrib['Target'] == part_name:
                        rel_root.remove(child)
            rels_part_name = rqet.rels_part_name_for_part(part_name)
        model.rels_forest.pop(rels_part_name)
    _remove_part_from_main_tree(model, part_name)
    model.set_modified()


def _duplicate_node(model, existing_node, add_as_part = True):
    """Creates a deep copy of the xml node (typically from another model) and optionally adds as part."""

    new_node = copy.deepcopy(existing_node)
    if add_as_part:
        uuid = rqet.uuid_for_part_root(new_node)
        if model.part_for_uuid(uuid) is None:
            _add_part(model, rqet.node_type(new_node), uuid, new_node)
    return new_node


def _force_consolidation_uuid_equivalence(model, immigrant_uuid, resident_uuid):
    """Force immigrant object to be teated as equivalent to resident during consolidation."""

    if model.consolidation is None:
        model.consolidation = cons.Consolidation(model)
    model.consolidation.force_uuid_equivalence(immigrant_uuid, resident_uuid)


def _copy_part_from_other_model(model,
                                other_model,
                                part,
                                realization = None,
                                consolidate = True,
                                force = False,
                                cut_refs_to_uuids = None,
                                cut_node_types = None,
                                self_h5_file_name = None,
                                h5_uuid = None,
                                other_h5_file_name = None):
    """Fully copies part in from another model, with referenced parts, hdf5 data and relationships."""

    # todo: double check behaviour around equivalent CRSes, especially any default crs in model

    assert other_model is not None
    if other_model is model:
        return part
    assert part is not None
    if realization is not None:
        assert isinstance(realization, int) and realization >= 0
    if force:
        assert consolidate
    if not other_h5_file_name:
        other_h5_file_name = other_model.h5_file_name()
    if not self_h5_file_name:
        self_h5_file_name = model.h5_file_name(file_must_exist = False)
    hdf5_copy_needed = not os.path.samefile(self_h5_file_name, other_h5_file_name)

    # check whether already existing in this model
    if part in model.parts_forest.keys():
        return part

    if other_model.type_of_part(part) == 'obj_EpcExternalPartReference':
        # log.debug('refusing to copy hdf5 ext part from other model')
        return None

    # log.debug('copying part: ' + str(part))

    uuid = rqet.uuid_in_part_name(part)
    if not force:
        assert model.part_for_uuid(uuid) is None, 'part copying failure: uuid exists for different part!'

    # duplicate xml tree and add as a part
    other_root = other_model.root_for_part(part, is_rels = False)
    if other_root is None:
        log.error('failed to copy part (missing in source model?): ' + str(part))
        return None

    resident_uuid = _unforced_consolidation(model, other_model, consolidate, force, part)

    if resident_uuid is None:

        root_node = _duplicate_node(model, other_root)  # adds duplicated node as part
        assert root_node is not None

        _set_realization_index_node_if_required(realization, root_node)

        if hdf5_copy_needed:
            # copy hdf5 data
            hdf5_internal_paths = [node.text for node in rqet.list_of_descendant_tag(other_root, 'PathInHdfFile')]
            hdf5_count = whdf5.copy_h5_path_list(other_h5_file_name, self_h5_file_name, hdf5_internal_paths, mode = 'a')
            # create relationship with hdf5 if needed and modify h5 file uuid in xml references
            _copy_part_hdf5_setup(model, hdf5_count, h5_uuid, root_node)
        # NB. assumes ext part is already established when sharing a common hdf5 file

        # cut references to objects to be excluded
        if cut_refs_to_uuids:
            rqet.cut_obj_references(root_node, cut_refs_to_uuids)

        if cut_node_types:
            rqet.cut_nodes_of_types(root_node, cut_node_types)

        # recursively copy in referenced parts where they don't already exist in this model
        _copy_referenced_parts(model, other_model, realization, consolidate, force, cut_refs_to_uuids, cut_node_types,
                               self_h5_file_name, h5_uuid, other_h5_file_name, root_node)

        resident_uuid = uuid

    else:

        root_node = model.root_for_uuid(resident_uuid)

    # copy relationships where target part is present in this model – this part is source, then destination
    _copy_relationships_for_present_targets(model, other_model, consolidate, force, resident_uuid, root_node)

    return model.part_for_uuid(resident_uuid)


def _unforced_consolidation(model, other_model, consolidate, force, part):
    if consolidate and not force:
        if model.consolidation is None:
            model.consolidation = cons.Consolidation(model)
        resident_uuid = model.consolidation.equivalent_uuid_for_part(part, immigrant_model = other_model)
    else:
        resident_uuid = None
    return resident_uuid


def _set_realization_index_node_if_required(realization, root_node):
    if realization is not None and rqet.node_type(root_node).endswith('Property'):
        ri_node = rqet.find_tag(root_node, 'RealizationIndex')
        if ri_node is None:
            ri_node = rqet.SubElement(root_node, ns['resqml2'] + 'RealizationIndex')
            ri_node.set(ns['xsi'] + 'type', ns['xsd'] + 'nonNegativeInteger')
        # NB. this intentionally overwrites any pre-existing realization number
        ri_node.text = str(realization)


def _copy_part_hdf5_setup(model, hdf5_count, h5_uuid, root_node):
    if hdf5_count:
        if h5_uuid is None:
            h5_uuid = model.h5_uuid()
        if h5_uuid is None:
            model.create_hdf5_ext()
            h5_uuid = model.h5_uuid()
        model.change_hdf5_uuid_in_hdf5_references(root_node, None, h5_uuid)
        ext_part = rqet.part_name_for_object('obj_EpcExternalPartReference', h5_uuid, prefixed = False)
        ext_node = model.root_for_part(ext_part)
        model.create_reciprocal_relationship(root_node,
                                             'mlToExternalPartProxy',
                                             ext_node,
                                             'externalPartProxyToMl',
                                             avoid_duplicates = False)


def _copy_referenced_parts(model, other_model, realization, consolidate, force, cut_refs_to_uuids, cut_node_types,
                           self_h5_file_name, h5_uuid, other_h5_file_name, root_node):
    for ref_node in rqet.list_obj_references(root_node):
        resident_referred_node = None
        if consolidate:
            resident_referred_node = model.referenced_node(ref_node, consolidate = True)
        if force:
            continue
        if resident_referred_node is None:
            referred_node = other_model.referenced_node(ref_node)
            if referred_node is None:
                log.warning('referred node not found in other model for ' +
                            f'{rqet.find_tag_text(ref_node, "Title")}; ' +
                            f'uuid: {rqet.find_tag_text(ref_node, "UUID")}')
            else:
                referred_part = rqet.part_name_for_part_root(referred_node)
                if other_model.type_of_part(referred_part) == 'obj_EpcExternalPartReference':
                    continue
                if referred_part in model.list_of_parts():
                    continue
                _copy_part_from_other_model(model,
                                            other_model,
                                            referred_part,
                                            realization = realization,
                                            consolidate = consolidate,
                                            force = force,
                                            cut_refs_to_uuids = cut_refs_to_uuids,
                                            cut_node_types = cut_node_types,
                                            self_h5_file_name = self_h5_file_name,
                                            h5_uuid = h5_uuid,
                                            other_h5_file_name = other_h5_file_name)


def _copy_relationships_for_present_targets(model, other_model, consolidate, force, resident_uuid, root_node):
    for source_flag in [True, False]:
        other_related_parts = other_model.parts_list_filtered_by_related_uuid(other_model.list_of_parts(),
                                                                              resident_uuid,
                                                                              uuid_is_source = source_flag)
        for related_part in other_related_parts:
            # log.debug('considering relationship with: ' + str(related_part))
            if not force and (related_part in model.parts_forest):
                resident_related_part = related_part
            else:
                # log.warning('skipping relationship between ' + str(part) + ' and ' + str(related_part))
                if consolidate:
                    resident_related_uuid = model.consolidation.equivalent_uuid_for_part(related_part,
                                                                                         immigrant_model = other_model)
                    if resident_related_uuid is None:
                        continue
                    resident_related_part = rqet.part_name_for_object(other_model.type_of_part(related_part),
                                                                      resident_related_uuid)
                    if resident_related_part is None:
                        continue
                else:
                    continue
            if not force and resident_related_part in model.parts_list_filtered_by_related_uuid(
                    model.list_of_parts(), resident_uuid):
                continue
            related_node = model.root_for_part(resident_related_part)
            assert related_node is not None

            if source_flag:
                sd_a, sd_b = 'sourceObject', 'destinationObject'
            else:
                sd_b, sd_a = 'sourceObject', 'destinationObject'
            model.create_reciprocal_relationship(root_node, sd_a, related_node, sd_b)


def _copy_all_parts_from_other_model(model, other_model, realization = None, consolidate = True):
    """Fully copies parts in from another model, with referenced parts, hdf5 data and relationships."""

    assert other_model is not None and other_model is not model

    other_parts_list = other_model.parts()
    if not other_parts_list:
        log.warning('no parts found in other model for merging')
        return

    if consolidate:
        other_parts_list = cons.sort_parts_list(other_model, other_parts_list)

    self_h5_file_name = model.h5_file_name(file_must_exist = False)
    self_h5_uuid = model.h5_uuid()
    other_h5_file_name = other_model.h5_file_name()
    for part in other_parts_list:
        _copy_part_from_other_model(model,
                                    other_model,
                                    part,
                                    realization = realization,
                                    consolidate = consolidate,
                                    self_h5_file_name = self_h5_file_name,
                                    h5_uuid = self_h5_uuid,
                                    other_h5_file_name = other_h5_file_name)

    if consolidate and model.consolidation is not None:
        model.consolidation.check_map_integrity()


def _set_uuid_to_part(model, part_name):
    """Adds an entry to the dictionary mapping from uuid to part name."""

    assert part_name and isinstance(part_name, str)
    uuid = rqet.uuid_in_part_name(part_name)
    model.uuid_part_dict[bu.uuid_as_int(uuid)] = part_name


def _del_uuid_to_part(model, part_name):
    """Deletes an entry from the dictionary mapping from uuid to part name."""

    uuid = rqet.uuid_in_part_name(part_name)
    try:
        del model.uuid_part_dict[bu.uuid_as_int(uuid)]
    except Exception:
        pass



"""Multiprocessing module containing the function used to run the wrapper functions in parallel."""

import logging
import time
from typing import List, Dict, Any, Callable, Union
from pathlib import Path
from resqpy.model import Model, new_model
from joblib import Parallel, delayed, parallel_backend  # type: ignore

log = logging.getLogger(__name__)


def rm_tree(path: Union[Path, str]) -> None:
    """Removes a directory using a pathlib Path.

    Args:
        path (Path/str): pathlib Path or string of the directory path.

    Returns:
        None.
    """
    path = Path(path)
    for child in path.iterdir():
        if child.is_file():
            child.unlink()
        else:
            rm_tree(child)
    path.rmdir()


def function_multiprocessing(function: Callable,
                             kwargs_list: List[Dict[str, Any]],
                             recombined_epc: Union[Path, str],
                             cluster,
                             consolidate: bool = True,
                             require_success = False) -> List[bool]:
    """Calls a function concurrently with the specfied arguments.

    A multiprocessing pool is used to call the function multiple times in parallel. Once
    all results are returned, they are combined into a single epc file.

    Args:
        function (Callable): the wrapper function to be called. Needs to return:

            - index (int): the index of the kwargs in the kwargs_list.
            - success (bool): whether the function call was successful, whatever that
                definiton is.
            - epc_file (Path/str): the epc file path where the objects are stored.
            - uuid_list (List[str]): list of UUIDs of relevant objects.

        kwargs_list (List[Dict[Any]]): A list of keyword argument dictionaries that are
            used when calling the function.
        recombined_epc (Path/str): A pathlib Path or path string of
            where the combined epc will be saved.
        cluster (LocalCluster/JobQueueCluster): a LocalCluster is a Dask cluster on a
            local machine. If using a job queing system, a JobQueueCluster can be used
            such as an SGECluster, SLURMCluster, PBSCluster, LSFCluster etc.
        consolidate (bool): if True and an equivalent part already exists in
            a model, it is not duplicated and the uuids are noted as equivalent.
        require_success (bool): if True and any instance fails, then an exception is
            raised

    Returns:
        success_list (List[bool]): A boolean list of successful function calls.

    Note:
        This function uses the Dask backend to run the given function in parallel, so a
        Dask cluster must be setup and passed as an argument. Dask will need to be
        installed in the Python environment because it is not a dependency of the
        project. More info can be found at 
        https://resqpy.readthedocs.io/en/latest/tutorial/multiprocessing.html
    """
    log.info("multiprocessing function called with %s function.", function.__name__)

    for i, kwargs in enumerate(kwargs_list):
        kwargs["index"] = i

    with parallel_backend("dask"):
        results = Parallel()(delayed(function)(**kwargs) for kwargs in kwargs_list)

    # Sorting the results by the original kwargs_list index.
    results = list(sorted(results, key = lambda x: x[0]))

    success_list = [result[1] for result in results]
    epc_list = [result[2] for result in results]
    uuids_list = [result[3] for result in results]
    success_count = sum(success_list)
    log.info("multiprocessing function calls complete; successes: %s/%s.", success_count, len(results))
    if require_success and success_count < len(results):
        raise Exception('one or more multiprocessing instances failed')

    epc_file = Path(str(recombined_epc))
    if epc_file.is_file():
        model_recombined = Model(epc_file = str(epc_file))
    else:
        model_recombined = new_model(epc_file = str(epc_file))

    log.info("creating the recombined epc file")
    for i, epc in enumerate(epc_list):
        if epc is None:
            continue
        while True:
            try:
                model = Model(epc_file = epc)
                break
            except FileNotFoundError:
                time.sleep(1)
                continue
        uuids = uuids_list[i]
        if uuids is None:
            uuids = model.uuids()
        for uuid in uuids:
            model_recombined.copy_uuid_from_other_model(model, uuid = uuid, consolidate = consolidate)

    # Deleting temporary directory.
    log.debug("deleting the temporary directory")
    rm_tree("tmp_dir")

    model_recombined.store_epc()
    model.h5_release()

    log.info("recombined epc file complete")

    return success_list



"""Units of measure."""

import json
import warnings
from functools import lru_cache
from pathlib import Path

from resqpy.olio.exceptions import IncompatibleUnitsError, InvalidUnitError

# physical constants
feet_to_metres = 0.3048
metres_to_feet = 1.0 / feet_to_metres
ft3_to_m3 = 0.028316846592
m3_to_ft3 = 1.0 / ft3_to_m3
bbl_to_m3 = 0.158987294928
m3_to_bbl = 1.0 / bbl_to_m3
psi_to_kPa = 44.482216152605 / 6.4516
kPa_to_psi = 1.0 / psi_to_kPa
d_to_s = float(24 * 60 * 60)
s_to_d = 1.0 / d_to_s

# Mapping from uom to set of common case-insensitive aliases
# Nb. No need to write out fractional combinations such as "bbl/day"
# note: some of the aliases are ambiguous, e.g. 'gm' could mean gigametre or gramme
UOM_ALIASES = {
    # Mass
    'g': {'gm', 'gram', 'gramme', 'grams', 'grammes'},
    'lbm': {'lb', 'lbs'},  # assumes pounds mass rather than pounds force

    # Length
    'm': {'m', 'metre', 'metres', 'meter', 'meters'},
    'ft': {'ft', 'foot', 'feet'},
    'cm': {'centimetre', 'centimetres', 'centimeter', 'centimeters'},

    # Time
    'ms': {'ms', 'msec', 'millisecs', 'millisecond', 'milliseconds'},
    's': {'s', 'sec', 'secs', 'second', 'seconds'},
    'min': {'min', 'mins', 'minute', 'minutes'},
    'h': {'h', 'hr', 'hour', 'hours'},
    'd': {'day', 'days'},
    'wk': {'wk', 'week', 'weeks'},
    'a': {'a', 'yr', 'year', 'years'},

    # Ratio
    '%': {'%', 'pu', 'p.u.', 'percent'},
    'm3/m3': {'m3/m3', 'v/v'},
    'g/cm3': {'g/cm3', 'g/cc'},

    # Volume
    'bbl': {'bbl', 'stb', 'rb'},
    '1000 bbl': {'1000 bbl', 'mstb', 'mbbl', 'mrb'},
    '1E6 bbl': {'1E6 bbl', 'mmstb', 'mmbbl'},
    '1E6 ft3': {'1E6 ft3', 'mmscf'},
    '1000 ft3': {'1000 ft3', 'mscf'},
    'm3': {'m3', 'sm3', 'stm3', 'rm3'},
    '1000 m3': {'kstm3', 'krm3', 'msm3', 'mrm3'},
    'ft3': {'ft3', 'scf', 'cf', 'rcf', 'cu.ft.'},
    'cm3': {'cc', 'scc', 'stcc'},
    'L': {'krcc', 'kstcc', 'kscc', 'litre', 'litres', 'liter', 'liters'},

    # Pressure & Reciprocal Pressure
    'psi': {'psi', 'psia'},

    # Thermodynamic Temperature
    'degC': {'c', 'degrees c', 'degreesc'},
    'degF': {'f', 'degrees f', 'degreesf'},

    # Energy
    'Btu[IT]': {'btu'},  # assumes BTU to refer to ISO standard, rather than older Btu[UK]
    'kJ': {'kj'},

    # Other
    'gAPI': {'gapi'},
    'S': {'mho'},
    'mS': {'mmho'},
    'mol': {'mole', 'moles'},
    'Euc': {'count', 'fraction', 'none'},
}
# Mapping from alias to valid uom
UOM_ALIAS_MAP = {alias.casefold(): uom for uom, aliases in UOM_ALIASES.items() for alias in aliases}

# Set of uoms that can be safely matched case-insensitive
CASE_INSENSITIVE_UOMS = {'m', 'ft', 'm3', 'ft3', 'm3/m3', 'ft3/ft3', 'bbl', 'bar', 'psi', 'm3/d', 'bbl/d'}


@lru_cache(None)
def rq_uom(units, quantity = None):
    """Returns RESQML uom string equivalent to units.

    Args:
       units (str): unit to coerce
       quantity (str, optional): if given, raise an exception if the uom is not supported
          for this quantity

    Returns:
       str: unit of measure

    Raises:
       InvalidUnitError: if units cannot be coerced into RESQML units for the given quantity
    """
    if not units:
        raise InvalidUnitError("Must provide non-empty unit")

    uom = _try_parse_unit(units.strip())

    if uom is None:
        raise InvalidUnitError(f"Cannot coerce {units} into a valid RESQML unit of measure.")

    if quantity is not None:
        supported_uoms = valid_uoms(quantity = quantity)
        if uom not in supported_uoms:
            raise InvalidUnitError(f"Unit {uom} is not supported for quantity {quantity}.\n"
                                   f"Supported units:\n{supported_uoms}")
    return uom


def convert(x, unit_from, unit_to, quantity = None, inplace = False):
    """Convert value between two compatible units.

    Args:
       x (numeric or np.array): value(s) to convert
       unit_from (str): resqml uom
       unit_to (str): resqml uom
       quantity (str, optional): If provided, raise an exception if units are not supported
          by this quantity
       inplace (bool): if True, convert arrays in-place. Else, return new value

    Returns:
       Converted value(s)

    Raises:
       InvalidUnitError: if units cannot be coerced into RESQML units
       IncompatibleUnitsError: if units do not have compatible base units
    """

    # conversion data assume the formula "y=(A + Bx)/(C + Dx)" where "y" represents a value in the base unit.
    # Backwards formula: x=(A-Cy)/(Dy-B)
    # All current units have D==0

    uom1 = rq_uom(unit_from, quantity = quantity)
    uom2 = rq_uom(unit_to, quantity = quantity)

    if uom1 == uom2:
        return x

    base1, dim1, (A1, B1, C1, D1) = get_conversion_factors(uom1)
    base2, dim2, (A2, B2, C2, D2) = get_conversion_factors(uom2)

    if base1 != base2:
        if dim1 != dim2:
            raise IncompatibleUnitsError(f"Cannot convert from '{unit_from}' to '{unit_to}':"
                                         f"\n - '{uom1}' has base unit '{base1} and dimension '{dim1}'."
                                         f"\n - '{uom2}' has base unit '{base2} and dimension '{dim2}'.")
        else:
            warnings.warn(f"Assuming base units {base1} and {base2} are equivalent as they have the same dimensions:"
                          f"\n - '{uom1}' has base unit '{base1} and dimension '{dim1}'."
                          f"\n - '{uom2}' has base unit '{base2} and dimension '{dim2}'.")

    if not inplace:
        y = (A1 + (B1 * x)) / (C1 + (D1 * x))
        return (A2 - (C2 * y)) / ((D2 * y) - B2)

    else:
        if any(f != 0 for f in [A1, A2, D1, D2]):
            raise NotImplementedError("In-place conversion not yet implemented for non-trivial conversions")

        factor = (B1 * C2) / (C1 * B2)
        x *= factor
        return x


@lru_cache(None)
def valid_uoms(quantity = None, return_attributes = False):
    """Return set of valid RESQML units of measure.

    Args:
       quantity (str): If given, filter to uoms supported by this quanitity.
       return_attributes (bool): If True, return a dict of all uoms and their
          attributes, such as the full name and dimension. Else, simply return
          the set of valid uoms.

    Returns
       set or dict
    """
    uoms = _properties_data()['units']
    if quantity:
        all_quantities = _properties_data()['quantities']
        supported_members = all_quantities[quantity]['members']
        uoms = {k: v for k, v in uoms.items() if k in supported_members}
    if return_attributes:
        return uoms
    else:
        return set(uoms.keys())


@lru_cache(None)
def valid_quantities(return_attributes = False):
    """Return set of valid RESQML quantities.

    Args:
       return_attributes (bool): If True, return a dict of all quantities and their
          attributes, such as the supported units of measure. Else, simply return
          the set of valid properties.

    Returns
       set or dict
    """
    quantities = _properties_data()['quantities']
    if return_attributes:
        return quantities
    else:
        return set(quantities.keys())


def valid_property_kinds():
    """Return set of valid property kinds."""

    return set(_properties_data()['property_kinds'].keys())


def rq_uom_list(units_list):
    """Returns a list of RESQML uom equivalents for units in list."""

    return [rq_uom(u) for u in units_list]


def rq_length_unit(units):
    """Returns length units string as expected by resqml."""

    return rq_uom(units, quantity = 'length')


def rq_time_unit(units):
    """Returns time units string as expected by resqml."""

    return rq_uom(units, quantity = 'time')


def convert_times(a, from_units, to_units, invert = False):
    """Converts values in numpy array (or a scalar) from one time unit to another, in situ if array.

    note:
       To see supported units, use: `valid_uoms(quantity='time')`
    """

    if invert:
        from_units, to_units = to_units, from_units

    return convert(a, from_units, to_units, quantity = 'time', inplace = True)


def convert_lengths(a, from_units, to_units):
    """Converts values in numpy array (or a scalar) from one length unit to another, in situ if array.

    arguments:
       a (numpy float array, or float): array of length values to undergo unit conversion in situ, or a scalar
       from_units (string): the units of the data before conversion
       to_units (string): the required units

    returns:
       a after unit conversion

    note:
       To see supported units, use: `valid_uoms(quantity='length')`
    """

    return convert(a, from_units, to_units, quantity = 'length', inplace = True)


def convert_pressures(a, from_units, to_units):
    """Converts values in numpy array (or a scalar) from one pressure unit to another, in situ if array.

    arguments:
       a (numpy float array, or float): array of pressure values to undergo unit conversion in situ, or a scalar
       from_units (string): the units of the data before conversion
       to_units (string): the required units

    returns:
       a after unit conversion

    note:
       To see supported units, use: `valid_uoms(quantity='pressure')`
    """
    return convert(a, from_units, to_units, quantity = 'pressure', inplace = True)


def convert_volumes(a, from_units, to_units):
    """Converts values in numpy array (or a scalar) from one volume unit to another, in situ if array.

    arguments:
       a (numpy float array, or float): array of volume values to undergo unit conversion in situ, or a scalar
       from_units (string): units of the data before conversion; see note for accepted units
       to_units (string): the required units; see note for accepted units

    returns:
       a after unit conversion

    note:
       To see supported units, use: `valid_uoms(quantity='volume')`
    """
    return convert(a, from_units, to_units, quantity = 'volume', inplace = True)


def convert_flow_rates(a, from_units, to_units):
    """Converts values in numpy array (or a scalar) from one volume flow rate unit to another, in situ if array.

    arguments:
       a (numpy float array, or float): array of volume flow rate values to undergo unit conversion in situ, or a scalar
       from_units (string): units of the data before conversion, eg. 'm3/d'; see notes for acceptable units
       to_units (string): required units of the data after conversion, eg. 'ft3/d'; see notes for acceptable units

    returns:
       a after unit conversion

    note:
       To see supported units, use: `valid_uoms(quantity='volume per time')`
    """
    return convert(a, from_units, to_units, quantity = 'volume per time', inplace = True)


@lru_cache(None)
def get_conversion_factors(uom):
    """Return base unit and conversion factors (A, B, C, D) for a given uom.

    The formula "y=(A + Bx)/(C + Dx)" where "y" represents a value in the base unit.

    Returns:
       3-tuple of (base_unit, dimension, factors). Factors is a 4-tuple of conversion factors

    Raises:
       ValueError if either uom is not a valid resqml uom
    """
    if uom not in valid_uoms():
        raise ValueError(f"{uom} is not a valid uom")
    uoms_data = _properties_data()["units"][uom]

    dimension = uoms_data["dimension"]
    try:
        a, b, c, d = uoms_data["A"], uoms_data["B"], uoms_data["C"], uoms_data["D"]
        base_unit = uoms_data["baseUnit"]
    except KeyError:  # Base units do not have factors defined
        a, b, c, d = 0, 1, 1, 0
        base_unit = uom
    return base_unit, dimension, (a, b, c, d)


# Private functions


@lru_cache(None)
def _properties_data():
    """Return a data structure that represents resqml unit system.

    The dict is loaded directly from a JSON file which is bundled with resqpy.
    The unit system is represented as a dict with the following keys:

    - dimensions
    - quantities
    - units
    - prefixes
    - property_kinds

    Returns:
       dict: resqml unit system
    """
    json_path = Path(__file__).parent / 'olio/data/properties.json'
    with open(json_path) as f:
        data = json.load(f)
    return data


def _try_parse_unit(units):
    """Try to match unit against known uoms and aliases, else return None."""

    uom_list = valid_uoms()
    ul = units.casefold()

    uom = None

    if units in uom_list:
        uom = units
    elif ul in CASE_INSENSITIVE_UOMS:
        uom = ul
    elif ul in UOM_ALIAS_MAP:
        uom = UOM_ALIAS_MAP[ul]
    elif ul in uom_list:
        uom = ul  # dangerous! for example, 'D' means D'Arcy and 'd' means day
    elif units.startswith('(') and units.endswith(')') and '(' not in units[1:]:  # simplistic
        uom = _try_parse_unit(units[1:-1])
    elif '/' in units:  # May be a fraction: match each part against known aliases
        parts = units.split('/', 1)
        newpart0 = _try_parse_unit(parts[0])
        newpart1 = _try_parse_unit(parts[1])
        if newpart0 and newpart1:
            ratio = f"{newpart0}/{newpart1}"
            if ratio in uom_list:
                uom = ratio

    return uom



"""Grid Connection Set class and related functions."""

__all__ = [
    'GridConnectionSet', 'pinchout_connection_set', 'k_gap_connection_set', 'cell_set_skin_connection_set',
    'add_connection_set_and_tmults', 'grid_columns_property_from_gcs_property', 'zero_base_cell_indices_in_faces_df',
    'standardize_face_indicator_in_faces_df', 'remove_external_faces_from_faces_df'
]

from ._grid_connection_set import GridConnectionSet
from ._gcs_functions import pinchout_connection_set, k_gap_connection_set,  \
    cell_set_skin_connection_set, add_connection_set_and_tmults,  \
    grid_columns_property_from_gcs_property, zero_base_cell_indices_in_faces_df,  \
    standardize_face_indicator_in_faces_df, remove_external_faces_from_faces_df

# Set "module" attribute of all public objects to this path.
for _name in __all__:
    _obj = eval(_name)
    if hasattr(_obj, "__module__"):
        _obj.__module__ = __name__



"""High level add_one_blocked_well_property() function."""

import os

import resqpy.model as rq
import resqpy.property as rqp
import resqpy.well as rqw


def add_one_blocked_well_property(epc_file,
                                  a,
                                  property_kind,
                                  blocked_well_uuid,
                                  source_info = 'imported',
                                  title = None,
                                  discrete = False,
                                  uom = None,
                                  time_index = None,
                                  time_series_uuid = None,
                                  string_lookup_uuid = None,
                                  null_value = None,
                                  indexable_element = 'cells',
                                  facet_type = None,
                                  facet = None,
                                  realization = None,
                                  local_property_kind_uuid = None,
                                  count_per_element = 1,
                                  points = False,
                                  extra_metadata = {},
                                  new_epc_file = None):
    """Adds a blocked well property from a numpy array to an existing resqml dataset.

    arguments:
       epc_file (string): file name to load model resqml model from (and rewrite to if new_epc_file is None)
       a (1D numpy array): the blocked well property array to be added to the model
       property_kind (string): the resqml property kind
       blocked_well_uuid (uuid object or string): the uuid of the blocked well to which the property relates
       source_info (string): typically the name of a file from which the array has been read but can be any
          information regarding the source of the data
       title (string): this will be used as the citation title when a part is generated for the array
       discrete (boolean, default False): if True, the array should contain integer (or boolean) data; if False, float
       uom (string, default None): the resqml units of measure for the data; not relevant to discrete data
       time_index (integer, default None): if not None, the time index to be used when creating a part for the array
       time_series_uuid (uuid object or string, default None): required if time_index is not None
       string_lookup_uuid (uuid object or string, optional): required if the array is to be stored as a categorical
          property; set to None for non-categorical discrete data; only relevant if discrete is True
       null_value (int, default None): if present, this is used in the metadata to indicate that this value
          is to be interpreted as a null value wherever it appears in the data (use for discrete data only)
       indexable_element (string, default 'cells'): the indexable element in the supporting representation (the blocked well);
          valid values are 'cells', 'intervals' (which includes unblocked intervals), or 'nodes'
       facet_type (string): resqml facet type, or None
       facet (string): resqml facet, or None
       realization (int): realization number, or None
       local_property_kind_uuid (uuid.UUID or string): uuid of local property kind, or None
       count_per_element (int, default 1): the number of values per indexable element; if greater than one then this
          must be the fastest cycling axis in the cached array, ie last index; if greater than 1 then a must be a 2D array
       points (bool, default False): if True, this is a points property with an extra dimension of extent 3
       extra_metadata (dict, optional): any items in this dictionary are added as extra metadata to the new
          property
       new_epc_file (string, optional): if None, the source epc_file is extended with the new property object; if present,
          a new epc file (& associated h5 file) is created to contain a copy of the blocked well (and dependencies) and
          the new property

    returns:
       uuid.UUID of newly created property object
    """

    if new_epc_file and epc_file and (
        (new_epc_file == epc_file) or
        (os.path.exists(new_epc_file) and os.path.exists(epc_file) and os.path.samefile(new_epc_file, epc_file))):
        new_epc_file = None

    # open up model and establish grid object
    model = rq.Model(epc_file)

    blocked_well = rqw.BlockedWell(model, uuid = blocked_well_uuid)
    assert blocked_well is not None, f'no blocked well object found with uuid {blocked_well_uuid}'

    if not discrete:
        string_lookup_uuid = None

    # create an empty property collection and add the new array to its 'imported' list
    bwpc = rqp.PropertyCollection()
    bwpc.set_support(support = blocked_well, model = model)
    bwpc.add_cached_array_to_imported_list(a,
                                           source_info,
                                           title,
                                           discrete = discrete,
                                           uom = uom,
                                           time_index = time_index,
                                           null_value = null_value,
                                           property_kind = property_kind,
                                           local_property_kind_uuid = local_property_kind_uuid,
                                           facet_type = facet_type,
                                           facet = facet,
                                           realization = realization,
                                           indexable_element = indexable_element,
                                           count = count_per_element,
                                           points = points)
    bwpc.write_hdf5_for_imported_list()
    uuid_list = bwpc.create_xml_for_imported_list_and_add_parts_to_model(time_series_uuid = time_series_uuid,
                                                                         string_lookup_uuid = string_lookup_uuid,
                                                                         property_kind_uuid = local_property_kind_uuid,
                                                                         extra_metadata = extra_metadata)
    assert len(uuid_list) == 1
    model.store_epc()
    return uuid_list[0]



"""High level add single cell grid function."""

import os
import numpy as np

import resqpy.olio.grid_functions as gf
import resqpy.rq_import as rqi


def add_single_cell_grid(points, new_grid_title = None, new_epc_file = None):
    """Creates a model with a single cell IJK Grid, with a cuboid cell aligned with x,y,z axes, enclosing points."""

    assert new_epc_file is not None

    # determine range of points
    min_xyz = np.nanmin(points.reshape((-1, 3)), axis = 0)
    max_xyz = np.nanmax(points.reshape((-1, 3)), axis = 0)
    assert not np.any(np.isnan(min_xyz)) and not np.any(np.isnan(max_xyz))

    # create corner point array in pagoda protocol
    cp = np.array([[min_xyz[0], min_xyz[1], min_xyz[2]], [max_xyz[0], min_xyz[1], min_xyz[2]],
                   [min_xyz[0], max_xyz[1], min_xyz[2]], [max_xyz[0], max_xyz[1], min_xyz[2]],
                   [min_xyz[0], min_xyz[1], max_xyz[2]], [max_xyz[0], min_xyz[1], max_xyz[2]],
                   [min_xyz[0], max_xyz[1], max_xyz[2]], [max_xyz[0], max_xyz[1], max_xyz[2]]]).reshape(
                       (1, 1, 1, 2, 2, 2, 3))

    # switch to nexus ordering
    gf.resequence_nexus_corp(cp)

    # write cp to temp pure binary file
    temp_file = new_epc_file[:-4] + '.temp.db'
    with open(temp_file, 'wb') as fp:
        fp.write(cp.data)

    # use_rq_import to create a new model
    one_cell_model = rqi.import_nexus(new_epc_file[:-4],
                                      extent_ijk = (1, 1, 1),
                                      corp_file = temp_file,
                                      ijk_handedness = 'left',
                                      use_binary = True,
                                      split_pillars = False,
                                      grid_title = new_grid_title)
    grid = one_cell_model.grid()

    os.remove(temp_file)

    return grid
